/* $Id: 542x_text16.inc,v 1.6 1998/11/24 18:28:09 marcus Exp $
******************************************************************************

   IBM (VGA) text16 operators

   Copyright (C) 1997 Jason McMullan	[jmcc@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

	This code was derived from the following source of information:
	[1]     Xfree 86 VGA driver
	[2]     Zen of Graphics Programming (Michael Abrash)

******************************************************************************
*/

#ifndef _VGA_TEXT16_INC_
#define _VGA_TEXT16_INC_

#include "IBM/vga_text16.h"

#define POINTER_MASK    0xFF		/* Attribute, Character */
#define CURSOR_MASK	0xE8		/* Attribute, Character */

typedef struct text16_point_s 
{
	int		visible;	/* Should I visible the old data? */
	int		offset;		/* Offset of old point */
	charcell	data;		/* Data at old point */
} text16_point;
	
struct text16_data_s 
{
	kgi_uint		magic;
	unsigned long	origin;
	int		splitline;
	text16_point	pointer;
#ifndef CURSOR_HW
	text16_point	cursor;
#endif
};

static struct text16_data_s text16_data[MAX_NR_DISPLAYS];

#define ORIGIN 		(data->origin)
#define SPLITLINE	(data->splitline)

#define PVISIBLE	(data->pointer.visible)
#define POFFSET		(data->pointer.offset)
#define PDATA		(data->pointer.data)

#ifndef CURSOR_HW
#define CVISIBLE	(data->cursor.visible)
#define COFFSET		(data->cursor.offset)
#define CDATA		(data->cursor.data)
#endif

#define GET_TEXT_DATA struct text16_data_s *data = \
		(struct text16_data_s *)KGIM_PRIVATE(dpy, driver)

static void text16_data_init(void)
{
	memset(text16_data, 0, sizeof(struct text16_data *) * MAX_NR_DISPLAYS);
}
	
static struct text16_data_s *text16_data_realloc(struct text16_data_s *old)
{
	int i;

	if (old == NULL || old->magic != TEXT16_MAGIC) {
		for (i = 0; i < MAX_NR_DISPLAYS; i++) 
			if (text16_data[i].magic != TEXT16_MAGIC) 
		{
			memset(&(text16_data[i]), 0, 
			       sizeof(struct text16_data_s));
			text16_data[i].magic = TEXT16_MAGIC;
			return &(text16_data[i]);
		}
		return NULL;
	}
	memset(old, 0, sizeof(struct text16_data_s));
	old->magic = TEXT16_MAGIC;
	return old;
}

static void text16_set_font(struct kgi_display *dpy, struct kgi_font *font,
			    kgi_uint from, kgi_uint to)
{
	IS_CRITICAL;
	kgiu8 SR2, SR4, GR4, GR5, GR6;
	int i,no=0;
	char *fbuf, *data;
	register int cnt = 32 - font->dy;

	if (font == NULL)
		return;

	fbuf = (char *) phys_to_virt(0x000A0000);
	fbuf += (no * 0x2000) + (from << 5);
	data = font->data + font->dy * from;
	if (!data)
		return;

	WAIT_VRETRACE();
	ENTER_CRITICAL;
	SR2 = inbSEQ(0x02);
	SR4 = inbSEQ(0x04);		/* Save register state        */
	GR4 = inbGRC(0x04);
	GR5 = inbGRC(0x05);
	GR6 = inbGRC(0x06);

	outbSEQ(SR02_PLANE2, 0x02);	/* Font data goes to plane 2  */
	outbSEQ(SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

	outbGRC(2, 0x04);		/* Read plane 2               */
	outbGRC(GR5_WRITEMODE0, 0x05);
	outbGRC(GR6_MAP_A0_64 | GR6_GRAPHMODE, 0x06);
	LEAVE_CRITICAL;
	
	/* NOTE: no printk()'s are allowed in here unless klogd is active! */

	for (i = from; i <= to; i++) {
		memmove(fbuf, data, font->dy);
		memset(fbuf + font->dy, 0, cnt);
		fbuf += 32;
		data += font->dy;
	};

	ENTER_CRITICAL;
	outbGRC(GR4, 0x04);
	outbGRC(GR5, 0x05);
	outbGRC(GR6, 0x06);
	outbSEQ(SR2, 0x02);
	outbSEQ(SR4, 0x04);		/* Restore registers  */
	LEAVE_CRITICAL;

#ifdef CURSOR_HW
	/* Set up cursor */
	outbCRT((inbCRT(0x0A) & ~CR0A_CURSOR_START_MASK)
		| ((font->base+1) & CR0A_CURSOR_START_MASK), 0x0A);
	outbCRT((inbCRT(0x0B) & ~CR0B_CURSOR_END_MASK)
		| ((font->base+2) & CR0B_CURSOR_END_MASK), 0x0B);
#endif
}

static void text16_set_origin(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	GET_TEXT_DATA;
	IS_CRITICAL;
	static kgi_uint last_xdot = -1;
	static kgi_uint last_ydot = -1;
	register kgi_uint xpix = x / dpy->mode.request.dpp.x;
	register kgi_uint xdot = x % dpy->mode.request.dpp.x;
	register kgi_uint ypix = y / dpy->mode.request.dpp.y;
	register kgi_uint ydot = y % dpy->mode.request.dpp.y;

	register kgi_uint offset = xpix + ypix * dpy->fb.size.x;

	DEBUG3("Origin: 0x%.8lx -> 0x%.8x", ORIGIN, offset);

	ENTER_CRITICAL;

	if (ORIGIN != offset) {

		outbCRT(offset >> 8, CRT_STARTADDR_H);
		outbCRT(offset, CRT_STARTADDR_L);
		ORIGIN = offset;
	}

	if (last_ydot != ydot) {

		outbCRT(ydot & CR08_ROWSCAN_MASK, CRT_PRESETROW);
		last_ydot = ydot;
	}

	if (last_xdot != xdot) {

		register kgi_uint hpan;

		if (dpy->mode.request.dpp.x == 9) {

			hpan = (xdot == 0) ? 0x08 : (xdot-1) & AR13_DATA_MASK;
		} else {

			hpan = xdot & AR13_DATA_MASK;
		}
		WAIT_HVRETRACE();
		outbATC(hpan, ATC_HPAN);
		outb( ATCI_ENABLEDISPLAY, ATCI);
		last_xdot = xdot;
	}

	LEAVE_CRITICAL;
}

static void text16_set_split(struct kgi_display *dpy, int line)
{
	GET_TEXT_DATA;
	register kgiu8 foo;

	outbCRT(line & 0xFF, CRT_LINECOMPARE);

	foo = (line & 0x0100)
		? inbCRT(0x07) | CR07_LINECOMPARE_B8
		: inbCRT(0x07) & ~CR07_LINECOMPARE_B8;
	outbCRT(foo, 0x07);

	foo = (line & 0x0200)
		? inbCRT(0x09) | CR09_LINECOMPARE_B9
		: inbCRT(0x09) & ~CR09_LINECOMPARE_B9;
	outbCRT(foo, 0x09);

	SPLITLINE=line;
}

#define	text16_set_frame	NULL	/* !!! implement! */

#define TEXTX           (dpy->mode.request.dpp.x)
#define TEXTY           (dpy->mode.request.dpp.y)
#define VIRTUALX        (dpy->fb.size.x)
#define VIRTUALY        (dpy->fb.size.y)
#define VISIBLEX        (dpy->mode.request.visible.x)
#define VISIBLEY        (dpy->mode.request.visible.y)

void text16_show_pointer(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	GET_TEXT_DATA;
	register kgiu32 newoff;
	register charcell *p = (void *) dpy->fb.pbuf;
	register charcell cell;

 	ASSERT((x < VISIBLEX) && (y < VISIBLEY));

 	if (SPLITLINE && y >= SPLITLINE) 
 		newoff = (x / TEXTX) + ((y-SPLITLINE) / TEXTY) * VIRTUALX;
 	else
 		newoff = ORIGIN + (x / TEXTX) + (y / TEXTY) * VIRTUALX;
 
	if (!PVISIBLE || newoff != POFFSET) {
		cell = *(p + newoff);
		ATTRSET(p + newoff, ATTR(cell) ^ POINTER_MASK);
	} else       
		return;

	if (PVISIBLE)
		*(p + POFFSET) = PDATA;
	else
		PVISIBLE = 1;

	POFFSET = newoff;
	PDATA = cell;
}

static void text16_hide_pointer(struct kgi_display *dpy)
{
	GET_TEXT_DATA;
	register charcell *p = (void *) dpy->fb.pbuf;

	if (PVISIBLE) {
		*(p + POFFSET) = PDATA;
		PVISIBLE = 0;
	}
}

#define text16_undo_pointer     text16_hide_pointer

#ifdef CURSOR_HW
static void text16_show_cursor(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	GET_TEXT_DATA;
	register kgiu32 new = ORIGIN + (x / TEXTX)  +  (y / TEXTY) * VIRTUALX;
 	IS_CRITICAL;

        ASSERT((x < VISIBLEX) && (y < VISIBLEY));

	ENTER_CRITICAL;
	outbCRT(new >> 8, CRT_CURSORADDR_H);
	outbCRT(new & 0xFF, CRT_CURSORADDR_L);
	LEAVE_CRITICAL;
}

static void text16_hide_cursor(struct kgi_display *dpy)
{
	IS_CRITICAL;

	ENTER_CRITICAL;
	outbCRT(0xFF, CRT_CURSORADDR_H);
	outbCRT(0xFF, CRT_CURSORADDR_L);
	LEAVE_CRITICAL;
}

#define text16_undo_cursor      text16_hide_cursor
#else
static void text16_show_cursor(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	GET_TEXT_DATA;
	register kgiu32 newoff;
	register kgiu16 *p = (void *) dpy->fb.pbuf;
	register kgiu16 cell;

	ASSERT((x < VISIBLEX) && (y < VISIBLEY));

 	if (SPLITLINE && y >= SPLITLINE) 
 		newoff = (x / TEXTX) + ((y-SPLITLINE) / TEXTY) * VIRTUALX;
 	else
 		newoff = ORIGIN + (x / TEXTX) + (y / TEXTY) * VIRTUALX;
 
	if (!CVISIBLE || newoff != COFFSET) {
		cell = *(p + newoff);
		ATTRSET(p + newoff, ATTR(cell) ^ CURSOR_MASK);
	} else 
		return;

	if (CVISIBLE)
		*(p + COFFSET) = CDATA;
	else
		CVISIBLE = 1;

	COFFSET=newoff;
	CDATA=cell;
}

static void text16_hide_cursor(struct kgi_display *dpy)
{
	GET_TEXT_DATA;
	register charcell *p = (void *) dpy->fb.pbuf;

	if (CVISIBLE)
		*(p + COFFSET) = CDATA;
	
	CVISIBLE = 0;
}

#define text16_undo_cursor      text16_hide_cursor
#endif /* CURSOR_HW */

#undef SPLITLINE
#undef ORIGIN

#undef PVISIBLE
#undef POFFSET
#undef PDATA

#ifndef CURSOR_HW
#undef CVISIBLE
#undef COFFSET
#undef CDATA
#endif

#endif /* _VGA_TEXT16_INC */
